use futures::future::join_all;
use reqwest;
use serde_json::{self, Value};
use std::{collections::HashSet, env, fs::File, io::Write};
use tl;

#[tokio::main]
async fn main() {
    let args: Vec<String> = env::args().collect();

    match args.len() {
        1 => {
            panic!("missing arguments");
        }
        2 => {
            let cmd = args[1].as_str();

            match cmd {
                "--all" => {
                    write_all().await;
                }
                "--help" => {
                    panic!("not implemented");
                }
                _ => {
                    panic!("bad command {}", cmd);
                }
            }
        }
        3 => {
            let cmd = args[1].as_str();
            let var = args[2].as_str();

            match cmd {
                "--item" => {
                    write_item(var.to_string()).await;
                }
                "--tier" => {
                    write_tier(var.to_string()).await;
                }
                "--test" => {
                    fetch_json(var.to_string()).await;
                }
                //"--milestone" => { }
                _ => {
                    panic!("bad command {}", cmd);
                }
            }
        }
        _ => {
            panic!("too many arguments");
        }
    }
}

async fn fetch_data(url: String) -> Vec<Value> {
    let webdata = reqwest::get(url)
        .await
        .expect("request failed")
        .text()
        .await
        .expect("request failed");

    let dom = tl::parse(webdata.as_str(), tl::ParserOptions::default()).expect("page is not html");
    let parser = dom.parser();
    let element = dom
        .get_element_by_id("__NUXT_DATA__")
        .expect("__NUXT_DATA__ not found")
        .get(parser)
        .unwrap();
    let data = element.inner_text(parser);
    serde_json::from_str(&data).unwrap()
}

async fn fetch_recipe_links(url: String) -> Vec<String> {
    let webdata = reqwest::get(url)
        .await
        .expect("request failed")
        .text()
        .await
        .expect("request failed");

    let dom = tl::parse(webdata.as_str(), tl::ParserOptions::default()).expect("page is not html");
    let parser = dom.parser();
    dom.query_selector("a") // Correctly query all <a> tags
        .expect("failed to query <a> tags")
        .filter_map(|e| {
            e.get(parser)
                .and_then(|node| node.as_tag()) // Ensure the node is a tag
                .and_then(|tag| tag.attributes().get("href")) // Get the href attribute
                .and_then(|attr| Some(attr?.as_utf8_str().to_string())) // Convert to String
        })
        .filter(|e| {
            e.contains("/show") && !e.contains("satisfactoryplus") && !e.contains("game-schematic")
        })
        .collect::<HashSet<_>>()
        .into_iter()
        .collect::<Vec<_>>()
}

async fn fetch_item(url: String) -> Value {
    let json = fetch_data(url.clone()).await;

    let mut v = json[0].clone();
    recurse(&mut v, &json);
    v[1]["data"]
        .as_object()
        .unwrap()
        .values()
        .next()
        .unwrap()
        .clone()
}
async fn write_item(url: String) {
    let result = fetch_item(url).await;
    let mut file = File::create(format!(
        "{}.json",
        &result["extraInformations"]["name"].as_str().unwrap()
    ))
    .unwrap();
    file.write(serde_json::to_string_pretty(&result).unwrap().as_bytes())
        .expect("failed to write json");
}

async fn fetch_tier(tier: String) -> Vec<Value> {
    let recipes =
        fetch_recipe_links(format!("https://wiki.kmods.space/milestones/tier-{}", tier)).await;

    join_all(
        recipes
            .iter()
            .map(|recipe| async move {
                fetch_item(format!("https://wiki.kmods.space{}/itemDescriptor", recipe)).await
            })
            .collect::<Vec<_>>(),
    )
    .await
}
async fn write_tier(tier: String) {
    let result = fetch_tier(tier.clone()).await;
    let mut file = File::create(format!("Tier {}.json", tier)).unwrap();
    file.write(serde_json::to_string_pretty(&result).unwrap().as_bytes())
        .expect("failed to write json");
}

async fn write_all() {
    let mut file = File::create("complete.json").unwrap();
    file.write(
        serde_json::to_string_pretty(
            &join_all(
                (0..9)
                    .into_iter()
                    .map(|i| async move { fetch_tier(i.to_string()).await })
            )
            .await
            .into_iter()
            .reduce(|mut acc, e| { acc.extend(e.iter().cloned()); acc })
        )
        .unwrap()
        .as_bytes(),
    )
    .unwrap();
}

async fn fetch_json(url: String) {
    let mut json = fetch_data(url).await;

    let mut v = json[0].clone();
    recurse(&mut v, &mut json);
    write_output("test.json", &v);
}

fn write_output(name: &str, content: &Value) {
    let mut file = File::create(name).unwrap();
    file.write(serde_json::to_string_pretty(&content).unwrap().as_bytes())
        .expect("failed to write json");
}

fn recurse<'a>(val: &'a mut Value, src: &'a Vec<Value>) {
    match val {
        Value::Array(arr) => {
            for val in arr.iter_mut() {
                recurse(val, src);
            }
        }
        Value::Object(map) => {
            for val in map.values_mut() {
                recurse(val, src);
            }
        }
        Value::Number(num) => {
            let i = num.as_u64().unwrap();
            let mut r = src.get(i as usize).unwrap().clone();
            match r {
                Value::Number(_) => {}
                _ => {
                    recurse(&mut r, src);
                }
            }
            *val = r;
        }
        _ => {}
    }
}
